﻿using System;
using System.Collections.Generic;
//using System.Linq;
using System.Text;
using System.Drawing;
using Tao.OpenGl;
using SdlDotNet.Graphics;
using SdlDotNet.Core;
using SdlDotNet.Input;

namespace SPS2
{
    public interface IScene
    {
        /// <summary>
        /// 描画を行う
        /// </summary>
        void Draw();

        /// <summary>
        /// フレーム処理を行う
        /// </summary>
        void Process();

        bool Enabled { get; set; }
        bool Visible { get; set; }
        bool Disposed { get; set; }

        void Focus();
        List<IScene> Children { get; }
        //void AddChild(IScene scene);
        //void RemoveChild(IScene scene);
        IScene Parent { get; set; }
        bool ProcessEvent(Tao.Sdl.Sdl.SDL_Event e);
    }

    public class SceneBase : IScene, IDisposable
    {
        public delegate void SceneEvent(IScene scene);
        public event SceneEvent DisposedEvent;
        private List<IScene> m_Children;
        private IScene m_Parent;

        public IScene Parent { get { return m_Parent; } set { m_Parent = value; } }

        private Display m_Display;
        protected Display Display { get { return m_Display; } }
        private bool m_Enabled;
        private bool m_Visible;
        private bool m_Disposed;
        public bool Enabled { get { return m_Enabled; } set { m_Enabled = value; } }
        public bool Visible { get { return m_Visible; } set { m_Visible = value; } }
        public bool Disposed { get { return m_Disposed; } set { m_Disposed = value; } }
        public SceneBase(Display d)
        {
            m_Display = d;
            m_Enabled = true;
            m_Visible = true;
            m_Disposed = false;
            m_Children = new List<IScene>();
        }
        public void Focus()
        {
            m_Display.Focus(this);
        }
        public virtual void Draw() { }
        public virtual void Process() { }
        public virtual void Dispose()
        {
            Disposed = true;
            if(DisposedEvent != null) DisposedEvent(this);
        }
        public List<IScene> Children { get { return m_Children; } }
        public virtual bool ProcessEvent(Tao.Sdl.Sdl.SDL_Event e) 
        {
            return false;
        }
        //public void AddChild(IScene scene)
        //{
        //    m_Children.Add(scene);
        //    m_Display.AddScene(scene);
        //    scene.Parent = this;
        //}
        //public void RemoveChild(IScene scene)
        //{
        //    m_Children.Remove(scene);
        //    m_Display.RemoveScene(scene);
 
        //    scene.Parent = null;
        //}
    }
    public class Scene_MouseAction : SceneBase
    {
        private class MotionUnit
        {
            public int X;
            public int Y;
            public DateTime TimeStamp;
            public MotionUnit(int x, int y)
            {
                X = x;
                Y = y;
                TimeStamp = DateTime.Now;
            }
        }
        private bool m_Mouse_Released;
        private Queue<MotionUnit> m_Motion;
        public Scene_MouseAction(Display d) : base(d) 
        {
            m_Motion = new Queue<MotionUnit>();
            m_Mouse_Released = false;
        }
        public override bool ProcessEvent(Tao.Sdl.Sdl.SDL_Event e)
        {
            if (e.type == Tao.Sdl.Sdl.SDL_MOUSEMOTION)
            {
                if ((e.motion.state & Tao.Sdl.Sdl.SDL_BUTTON_LEFT) > 0)
                {
                    if (m_Mouse_Released)
                    {
                        m_Motion.Enqueue(new MotionUnit(-1, -1));
                    }
                    m_Motion.Enqueue(new MotionUnit(e.motion.x, e.motion.y));
                    m_Mouse_Released = false;
                }
                else
                {
                    m_Mouse_Released = true;
                }
            }
            return false;   
        }
        public override void Process()
        {
            while (m_Motion.Count > 0)
            {
                MotionUnit u = m_Motion.Peek();
                if ((DateTime.Now - u.TimeStamp).TotalMilliseconds > 3000) m_Motion.Dequeue();
                else break;
            }
        }
        public override void Draw()
        {
            Gl.glDisable(Gl.GL_TEXTURE_2D);
            Gl.glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
            Gl.glLineWidth(3.0f);
            Gl.glBegin(Gl.GL_LINE_STRIP);
            foreach (MotionUnit p in m_Motion)
            {
                if (p.X == -1 && p.Y == -1)
                {
                    Gl.glEnd();
                    Gl.glBegin(Gl.GL_LINE_STRIP);
                }
                else
                {
                    Gl.glVertex2f(p.X - Display.Width / 2.0f, p.Y - Display.Height / 2.0f);
                }
            }
            Gl.glEnd();
            Gl.glEnable(Gl.GL_TEXTURE_2D);
            Gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f); // TODO:これやっとかないとFlipで困るんだけど美しくない
        }
    }
    public class Scene_PageSelector : SceneBase
    {
        static void PageIn(Transition t, Page p, long Elapsed_ms, int direction)
        {
            float alpha = t.ScaleFactor(0.0f, 0.6f);
            float angle = t.ScaleFactor(160.0f, 0.0f);
            float offset = t.ScaleFactor(t.Display.Height / 2, 0.0f); // TODO:このへんのパラメータは共有したい

            Gl.glPushMatrix();
            Gl.glTranslatef(direction * t.Display.Width / 4.0f + offset * direction, t.Display.Height / 4.0f - offset * direction, 0.0f);
            Gl.glScalef(0.4f, 0.4f, 1.0f);
            Gl.glRotatef(angle, 0.0f, 0.0f, 1.0f);
            Gl.glColor4f(1.0f, 1.0f, 1.0f, alpha);
            p.Draw(t.Display, 0);
            Gl.glPopMatrix();
        }
        static void PageOut(Transition t, Page p, long Elapsed_ms, int direction)
        {
            float alpha = t.ScaleFactor(0.6f, 0.0f);
            float angle = t.ScaleFactor(0.0f, 160.0f);
            float offset = t.ScaleFactor(0.0f, t.Display.Height / 2);

            Gl.glPushMatrix();
            Gl.glTranslatef(direction * t.Display.Width / 4.0f + offset * direction, t.Display.Height / 4.0f - offset * direction, 0.0f);
            Gl.glScalef(0.4f, 0.4f, 1.0f);
            Gl.glRotatef(angle, 0.0f, 0.0f, 1.0f);
            Gl.glColor4f(1.0f, 1.0f, 1.0f, alpha);
            p.Draw(t.Display, 0);
            Gl.glPopMatrix();
        }
        class Flip_Initial : Transition
        {
            public Flip_Initial(Display d, int pageno) : base(d, true, 400)
            {
                m_Pageno = pageno;
            }
            private int m_Pageno;

            public override void Draw()
            {
                int pageno = m_Pageno; // TODO:なんとかせい

                long tick = Elapsed_ms;
                float alpha = ScaleFactor(0.0f, 0.6f);
                float angle = ScaleFactor(160.0f, 0.0f);
                float offset = ScaleFactor(Display.Height / 2, 0.0f);

                PageIn(this, Display.presentation.PrevPage(pageno), Elapsed_ms, -1);
                PageIn(this, Display.presentation.NextPage(pageno), Elapsed_ms, 1);

                float scale = ScaleFactor(1.0f, 0.6f);
                Gl.glScalef(scale, scale, 1.0f);
                Page current = Display.presentation.Pages[pageno];
                Gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
                current.Draw(Display, 0);
            }
        }
        class Flip_Terminal : Transition
        {
            public Flip_Terminal(Display d, int pageno) : base(d, true, 400) 
            {
                m_Pageno = pageno;
            }
            private int m_Pageno;

            public override void Draw()
            {
                int pageno = m_Pageno; // TODO:なんとかせい

                long tick = Elapsed_ms;
                float alpha = ScaleFactor(0.6f, 0.0f);
                float angle = ScaleFactor(0.0f, 160.0f);
                float offset = ScaleFactor(0.0f, Display.Height / 2.0f); // TODO: ちょうど逆

                PageOut(this, Display.presentation.PrevPage(pageno), Elapsed_ms, -1);
                PageOut(this, Display.presentation.NextPage(pageno), Elapsed_ms, 1);

                float scale = ScaleFactor(0.6f, 1.0f);
                Gl.glScalef(scale, scale, 1.0f);
                
                Page current = Display.presentation.Pages[pageno];
                Gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
                current.Draw(Display, 0);
            }
        }
        class Flip_Selecting : Transition
        {
            public Flip_Selecting(Display d, int pageno, int direction)
                : base(d, true, 400)
            {
                m_Pageno = pageno;
                m_Direction = direction;
                System.Diagnostics.Debug.Assert(direction == 1 || direction == -1);
            }
            private int m_Pageno;
            private int m_Direction;
            public override void Draw()
            {
                int pageno = m_Pageno; // TODO:なんとかせい

                long tick = Elapsed_ms;
                float angle = ScaleFactor(160.0f, 0.0f);
                float fadein1 = ScaleFactor(0.0f, 0.6f);
                float fadein2 = ScaleFactor(0.6f, 1.0f);
                float fadeout1 = ScaleFactor(0.6f, 0.0f);
                float fadeout2 = ScaleFactor(1.0f, 0.6f);

                float offset_x = ScaleFactor(0.0f, Display.Width / 4.0f);
                float offset_y = ScaleFactor(0.0f, Display.Height / 4.0f);
                float zoomin = ScaleFactor(0.4f, 0.6f);
                float zoomout = ScaleFactor(0.6f, 0.4f);


                Page newcurrent = Display.presentation.NextPage(pageno);
                if (m_Direction == 1)
                {
                    newcurrent = Display.presentation.PrevPage(pageno);
                }
                // 次に選択されるページを移動
                Gl.glPushMatrix();
                Gl.glTranslatef(-m_Direction * Display.Width / 4.0f + offset_x * m_Direction, 
                    Display.Height / 4.0f - offset_y, 0.0f);
                Gl.glScalef(zoomin, zoomin, 1.0f);
                Gl.glColor4f(1.0f, 1.0f, 1.0f, fadein2);
                newcurrent.Draw(Display, 0);
                Gl.glPopMatrix();

                // 現在のページを隣に移動

                Gl.glPushMatrix();
                Gl.glTranslatef(offset_x * m_Direction, offset_y , 0.0f);
                Gl.glScalef(zoomout, zoomout, 1.0f);
                Page current = Display.presentation.Pages[pageno];
                Gl.glColor4f(1.0f, 1.0f, 1.0f, fadeout2);
                current.Draw(Display, 0);
                Gl.glPopMatrix();

                Page leave = Display.presentation.PrevPage(pageno);
                Page enter = Display.presentation.NextPage(Display.presentation.NextPageNo(pageno));
                if(m_Direction == 1)
                {
                    leave = Display.presentation.NextPage(pageno);
                    enter = Display.presentation.PrevPage(Display.presentation.PrevPageNo(pageno));
                }
                // 表示されなくなるページ
                PageOut(this, leave, Elapsed_ms, m_Direction);


                // 新しく表示されるページ
                PageIn(this, enter, Elapsed_ms, -m_Direction);

            }
        }
        public Scene_PageSelector(Display d)
            : base(d) 
        {
            m_Pageno = d.pageno;
            Visible = false;
            Scene_Transition f = new Scene_Transition(d, new Flip_Initial(d, m_Pageno)); 
            f.DisposedEvent += new SceneEvent(Flip_Finished);
            Display.AddScene(f);
            f.Focus();
        }
        private void Flip_Finished(IScene scene)
        {
            this.Visible = true;
        }
        private void Terminal_Finished(IScene scene)
        {
            Dispose();
        }
        private int m_Pageno;
        public override bool ProcessEvent(Tao.Sdl.Sdl.SDL_Event e)
        {
            if (e.type == Tao.Sdl.Sdl.SDL_KEYDOWN)
            {
                switch (e.key.keysym.sym)
                {
                    case Tao.Sdl.Sdl.SDLK_RIGHT:
                        Select_Page(-1);
                        m_Pageno = Display.presentation.NextPageNo(m_Pageno);
                        break;
                    case Tao.Sdl.Sdl.SDLK_LEFT:
                        Select_Page(1);
                        m_Pageno = Display.presentation.PrevPageNo(m_Pageno);
                        break;
                    case Tao.Sdl.Sdl.SDLK_SPACE:
                    case Tao.Sdl.Sdl.SDLK_p:
                    case Tao.Sdl.Sdl.SDLK_RETURN:
                        Select_Complete();
                        break;
                }
                return true;
            }
            return false;   
        }
        private void Select_Page(int direction)
        {
            Visible = false;
            Scene_Transition f = new Scene_Transition(Display, new Flip_Selecting(Display, m_Pageno, direction));
            f.DisposedEvent += new SceneEvent(Flip_Finished);
            Display.AddScene(f);
            f.Focus();

        }

        private void Select_Complete()
        {
            Display.pageno = m_Pageno;
            Visible = false;
            Scene_Transition f = new Scene_Transition(Display, new Flip_Terminal(Display, m_Pageno));
            f.DisposedEvent += new SceneEvent(Terminal_Finished);
            Display.AddScene(f); // TODO: Sceneという名前は変えたほうがよさそう。Windowか？
            f.Focus();
        }
        public override void Draw()
        {
            Gl.glColor4f(1.0f, 1.0f, 1.0f, 0.6f);
            Gl.glPushMatrix();
            Gl.glTranslatef(-Display.Width / 4.0f, Display.Height / 4.0f, 0.0f);
            Gl.glScalef(0.4f, 0.4f, 1.0f);
            Display.presentation.PrevPage(m_Pageno).Draw(Display, 0);
            Gl.glPopMatrix();

            Gl.glPushMatrix();
            Gl.glTranslatef(Display.Width / 4.0f, Display.Height / 4.0f, 0.0f);
            Gl.glScalef(0.4f, 0.4f, 1.0f);
            Display.presentation.NextPage(m_Pageno).Draw(Display, 0);
            Gl.glPopMatrix();

            Gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
            // 縮小表示
            Gl.glScalef(0.6f, 0.6f, 1.0f);
            Display.presentation.Pages[m_Pageno].Draw(Display, 0);
        }

    }
    public class Scene_Transition : SceneBase
    {
        private Transition m_Flip;
        public Scene_Transition(Display d, Transition f) : base(d) 
        {
            m_Flip = f;
        }
        public override bool ProcessEvent(Tao.Sdl.Sdl.SDL_Event e)
        {
            if (e.type == Tao.Sdl.Sdl.SDL_KEYDOWN)
            {
                return KeyDown();
            }
            return false;
        }
        private bool KeyDown()
        {
            // Flip中止
            m_Flip.Done = true;
            Dispose();
            return true;
        }
        public override void Draw()
        {
            m_Flip.Draw();
        }
        public override void Process()
        {
            m_Flip.Process();
            if (m_Flip.Done) Dispose();
        }
        public override void Dispose()
        {
            base.Dispose();
        }
    }
    public class Scene_Fps : SceneBase
    {
        private DateTime m_Last;
        private int m_Frames;
        private int m_Interval_ms = 800;
        private double m_Fps;
        private SurfaceGl m_Surface;
        public Scene_Fps(Display d) : base(d)
        {
            m_Last = DateTime.Now;
            m_Fps = 0.0d;
        }
        public override void Draw()
        {
            try
            {
                m_Frames++;
                TimeSpan diff = DateTime.Now - m_Last;
                if (diff.TotalMilliseconds > m_Interval_ms)
                {
                    m_Fps = (double)m_Frames * 1000.0 / (double)diff.TotalMilliseconds;

                    string message = m_Fps.ToString("0.0") + " fps";
                    m_Surface = new SurfaceGl(Display.SPS.Fonts["4"].Render(message, Color.Yellow, true)); // TODO:XXX

                    m_Last = DateTime.Now;
                    m_Frames = 0;
                }
                if (m_Surface == null) return;
                Gl.glColor4f(0.0f, 0.0f, 0.0f, 0.5f);

                Gl.glDisable(Gl.GL_TEXTURE_2D);
                Gl.glPushMatrix();
                Gl.glBegin(Gl.GL_QUADS);
                Gl.glVertex2f(Display.Width / 2.0f - m_Surface.Surface.Width, -Display.Height / 2.0f);
                Gl.glVertex2f(Display.Width / 2.0f, -Display.Height / 2.0f);
                Gl.glVertex2f(Display.Width / 2.0f, -Display.Height / 2.0f + m_Surface.Surface.Height);
                Gl.glVertex2f(Display.Width / 2.0f - m_Surface.Surface.Width, -Display.Height / 2.0f + m_Surface.Surface.Height);
                Gl.glEnd();
                Gl.glPopMatrix();
                Gl.glEnable(Gl.GL_TEXTURE_2D);
                Gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
                m_Surface.Draw(
                    Display.Width / 2.0f - m_Surface.Surface.Width,
                    -Display.Height / 2.0f);
            }
            finally
            {
                Gl.glColor4f(1.0f, 1.0f, 1.0f, 1.0f); // TODO:これやっとかないとFlipで困るんだけど美しくない
            }
            // TODO: TOPMOSTあるいはZorderが必要
        }
    }
    public class Scene_TopLevel : SceneBase
    {
        public Scene_TopLevel(Display d) : base(d)
        {
        }
        public override void Draw()
        {
            if (Display.Scene == null) return;
            Display.Scene.Draw(Display); // TODO:なんか不自然
        }
        public override bool ProcessEvent(Tao.Sdl.Sdl.SDL_Event e)
        {
            if (e.type != Tao.Sdl.Sdl.SDL_KEYDOWN)
            {
                return false;
            }
            else
            {
                return KeyDown(e.key);
            }
        }

        private bool KeyDown(Tao.Sdl.Sdl.SDL_KeyboardEvent e)
        {
            switch (e.keysym.sym)
            {
                case Tao.Sdl.Sdl.SDLK_SPACE:
                case Tao.Sdl.Sdl.SDLK_RIGHT: 
                    Next();
                    return true;
                case Tao.Sdl.Sdl.SDLK_BACKSPACE:
                case Tao.Sdl.Sdl.SDLK_LEFT:
                    Prev();
                    return true;
                case Tao.Sdl.Sdl.SDLK_p:
                    PageSelector();
                    return true;
            }
            return false;
        }
        private void Prev()
        {
            if (Display.Scene.pauselevel == 0)
            {
                // previous page
                if (Display.pageno == 0)
                {
                    Display.pageno = Display.presentation.pageno_max;
                }
                else
                {
                    Display.pageno--;
                }
            }
            else
            {
                Display.Scene.pauselevel--;
            }
        }
        private void Next()
        {
            if (Display.Scene.pauselevel == Display.Scene.page.pauselevel_max)
            {
                Flip.Flip_Parameter p = new Flip.Flip_Parameter();
                p.Scene_before = new Scene(Display.Scene);

                // next page
                if (Display.pageno == Display.presentation.pageno_max)
                {
                    Display.pageno = 0;
                }
                else
                {
                    Display.pageno++;
                }
                p.Scene_after = Display.Scene;

                Visible = false;
                Scene_Transition flip = new Scene_Transition(Display,
                    Flip.CreateInstance(p.Scene_after.page.setting.flip_driver, Display, p));
                flip.DisposedEvent += new SceneEvent(Flip_Disposed);
                Display.AddScene(flip);
                flip.Focus();
            }
            else
            {
                Display.Scene.pauselevel++;
            }
        }
        private void PageSelector()
        {
            Visible = false;
            Scene_PageSelector psel = new Scene_PageSelector(Display);
            psel.DisposedEvent += new SceneEvent(Flip_Disposed);
            Display.AddScene(psel);
            psel.Focus();
        }
        private void Flip_Disposed(IScene flip)
        {
            Visible = true;
        }

    }
    public class Display
    {
        private IScene m_TopScene;
        private List<IScene> m_Zorder;
        public List<IScene> Zorder_ { get { return m_Zorder; } } // TODO: 外に出さない
        public IScene TopScene_ { get { return m_TopScene; } }

        public void AddScene(IScene scene, IScene parent)
        {
            parent.Children.Add(scene);
            scene.Parent = parent;
            m_Zorder.Add(scene);
        }
        public void AddScene(IScene scene)
        {
            AddScene(scene, m_TopScene);
        }
        public void RemoveScene(IScene scene)
        {
            scene.Parent.Children.Remove(scene);
            scene.Parent = null;

            m_Zorder.Remove(scene);
        }
        public void Focus(IScene scene)
        {
            if (m_Zorder.Contains(scene))
            {
                m_Zorder.Remove(scene);
            }
            m_Zorder.Insert(0, scene);
        }

        public Surface screen;
        public Presentation presentation;
        private Scene m_Scene;
        public Scene Scene {
            get { return m_Scene; }
            set {
                m_Scene = value;
                m_TopScene = new Scene_TopLevel(this);
                m_Zorder = new List<IScene>();
                m_Zorder.Add(m_TopScene);
                m_TopScene.Focus();
            }
        }

        private SPS2 m_Sps;
        public SPS2 SPS { get { return m_Sps; } set { m_Sps = value; } }

        public void ProcessEvent(Tao.Sdl.Sdl.SDL_Event e)
        {
            foreach (IScene scene in m_Zorder)
            {
                if(scene.ProcessEvent(e)) break;
            }
        }
        public Display()
        {
            setting = new Setting();

        }
        private int m_pageno;
        public int pageno
        {
            get { return m_pageno; }
            set
            {
                Scene.page = presentation.Pages[value];
                Scene.pauselevel = 0;
                m_pageno = value;
            }
        }

        private void Draw(IScene scene)
        {
            if (scene.Visible) scene.Draw();
            foreach (IScene child in scene.Children)
            {
                Draw(child);
            }
        }
        private void Process(IScene scene)
        {
            List<IScene> delete = new List<IScene>();
            if (scene.Enabled) scene.Process();
            foreach (IScene child in scene.Children)
            {
                Process(child);
                if (child.Disposed)
                {
                    delete.Add(child);
                }
            }
            if (delete.Count == 0) return;
            foreach (IScene child in delete)
            {
                RemoveScene(child);
            }
        }
        public void Draw()
        {
            Draw(m_TopScene);
        }
        public void Process()
        {
            Process(m_TopScene);
        }

        public class Setting
        {
            public bool showfps;
            public bool fullscreen;
            public bool use_opengl;
        }
        public Setting setting;

        public int RatioToPixelH(double h)
        {
            return (int)(Width * h / 100.0);
        }
        public int RatioToPixelV(double v)
        {
            return (int)(Height * v / 100.0);
        }
        public int RatioToPixel(double l)
        {
            if (Width > Height)
            {
                return RatioToPixelV(l);
            }
            else
            {
                return RatioToPixelH(l);
            }
        }

        public int Width;
        public int Height;
    }

}
